from __future__ import annotations

from pathlib import Path
from typing import Union, Dict, Optional

import attr
from box import Box
from jinja2 import Template

import black
from ruamel import yaml


def yaml_to_box(filename: Union[Path, str]) -> Box:
    """
    Takes a yaml file name and returns a Box dictionary.

    :param filename: Union[Path, str]
    :return: Box
    """
    filename = Path(filename)
    if not filename.is_file() and filename.parent != Path(__file__).parent:
        return yaml_to_box(Path(__file__).parent / filename.name)
    return Box(yaml.safe_load(filename.open("r", encoding="utf8")))


@attr.s
class SchemaGen:
    """
    Main class of the program that does all the work. Once the code is generated the output is stored in
    :class:`SchemaGen.code` property. You can choose to either save it to file with :class:`SchemaGen.to_file`
    or import to sys with :class:`SchemaGen.to_sys`

    Example usage:
    ::
        from pydantic_gen import SchemaGen

        generated = SchemaGen('example.yml')

    Once the code is generated you can save it to file:
    ::
        generated.to_file('generated_schemas.py')

    Or make it directly importable by adding to `sys.modules`:
    ::
        generated.to_sys(module_name='generated_schemas')

    After running `generated.to_sys(module_name='generated_schemas'` your code will be available for import:
    ::
        import generated_schemas as gs
    """

    filename: Optional[Union[Path, str, None]] = attr.ib(default=None)
    yaml_content: Dict = attr.ib(default=None)
    code: str = attr.ib(default=None)
    _additional_imports = set()

    _templates: Box = yaml_to_box("templates.yml")
    _module: Template = Template(_templates.module)
    _schemas: Template = Template(_templates.schemas)
    _schema: Template = Template(_templates.schema)
    _prop: Template = Template(_templates.prop)
    _config: Template = Template(_templates.config)

    def __attrs_post_init__(self):
        if not self.yaml_content:
            self._content_from_file()
        self.code = self._make_module_and_schemas()
        self.code = black.format_str(self.code, mode=black.FileMode())

    def _content_from_file(self):
        self.filename = Path(self.filename)
        if not self.filename.is_file():
            raise FileNotFoundError(f"{self.filename.name}")
        self.yaml_content = yaml_to_box(self.filename)

    @classmethod
    def from_string(cls, s: str) -> SchemaGen:
        """
        Generates code from a yaml string instead of file.

        :param s: yaml string
        :return: SchemaGen
        """
        d = yaml.safe_load(s)
        return cls.from_dict(d)

    @classmethod
    def from_dict(cls, d: dict) -> SchemaGen:
        """
        Alternative constructor. Takes a dict from reading a yaml file instead of the file itself and constructs a
        :class:`SchemaGen` object.

        :param d: dict
        :return: SchemaGen
        """
        return cls(yaml_content=Box(d))

    def _make_module_and_schemas(self) -> str:
        schemas = self._schemas.render(schemas=self._make_schemas())
        additional_imports = (
            "\n".join(sorted(list(self._additional_imports)))
            if self._additional_imports
            else ""
        )
        module = self._module.render(imports=additional_imports)
        return "\n\n\n".join([module, schemas])

    def _make_schemas(self) -> str:
        schemas = []
        for schema in self.yaml_content.schemas:
            schemas.append(self._make_schema(schema))
        return "\n\n\n".join(schemas)

    def _make_schema(self, schema: Box) -> str:
        props = [self._make_prop(prop) for prop in schema.props]
        config = self._make_config(schema.get("config"))
        config_text = "\n\n" + 4 * " " + config if config else ""
        props_text = "\n    ".join(props)
        return self._schema.render(
            name=schema.name, props=props_text, config=config_text
        )

    def _make_prop(self, prop: Box) -> str:
        if not prop.get("default"):
            default_value = ""
        elif prop.type in ["str", "date", "datetime", "time"]:
            default_value = f' = "{prop.default}"'
        else:
            default_value = f" = {prop.default}"

        if prop.type in ["date", "datetime", "time"]:
            self._additional_imports.add("import datetime as dt")
            prop.type = f"dt.{prop.type}"

        if prop.type == "uuid":
            self._additional_imports.add("import uuid")
            prop.type = "uuid.UUID"

        if prop.get("optional"):
            prop_type = f"Optional[{prop.type}]"
        else:
            prop_type = prop.type
        return self._prop.render(name=prop.name, type=prop_type, default=default_value)

    def _make_config(self, config: Box) -> str:
        if not config:
            return ""
        else:
            conf_text = ""
            for conf in config:
                conf_text += (
                    4 * " " + f"{list(conf.keys())[0]} = {list(conf.values())[0]}\n"
                )
        return self._config.render(confs=conf_text)

    def to_file(self, filename: Union[Path, str]) -> None:
        """
        Saves generated code to file.

        :param filename: Union[Path, str]
        :return: None
        """
        filename = Path(filename)
        filename.write_text(self.code, encoding="utf8")
        return filename

    def to_sys(self, module_name: str) -> None:
        """
        Saves generated code into sys.modules making the classes available for import.

        :param module_name: str : name of the module under which the code will be importable
        :return: None

        Example:
        ::
            from pydantic_gen import SchemaGen

            generated = SchemaGen('example.yml')
            generated.to_sys(module_name='generated_schemas')

        After running `generated.to_sys(module_name='generated_schemas'` your code will be available for import:
        ::
            import generated_schemas as gs
        """

        import sys
        from types import ModuleType

        mod = ModuleType(module_name)
        sys.modules[module_name] = mod
        exec(self.code, mod.__dict__)

    def print(self):
        """
        Prints the generated code.
        """
        print(self.code)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydantic_gen']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3.0,<20.0.0',
 'black>=19.10b0,<20.0',
 'jinja2>=2.11.2,<3.0.0',
 'pydantic>=1.4,<2.0',
 'python-box>=4.2.2,<5.0.0',
 'ruamel.yaml>=0.16.10,<0.17.0']

setup_kwargs = {
    'name': 'pydantic-gen',
    'version': '0.3.3',
    'description': 'Code generator for pydantic schemas',
    'long_description': "==========================\nPydantic Schemas Generator\n==========================\n.. image:: https://img.shields.io/pypi/pyversions/pydantic-gen\n    :target: https://pypi.org/project/pydantic-gen/\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n.. image:: https://img.shields.io/pypi/v/pydantic-gen\n    :target: https://pypi.org/project/pydantic-gen/\n.. image:: https://img.shields.io/pypi/dw/pydantic-gen\n    :target: https://pypi.org/project/pydantic-gen/\n.. image:: https://img.shields.io/endpoint.svg?url=https%3A%2F%2Factions-badge.atrox.dev%2Flicht1stein%2Fpydantic-gen%2Fbadge&style=flat\n    :target: https://actions-badge.atrox.dev/licht1stein/pydantic-gen/goto\n.. image:: https://img.shields.io/readthedocs/pydantic-gen   :alt: Read the Docs\n    :target: https://pydantic-gen.readthedocs.io/en/latest/\n.. image:: https://img.shields.io/badge/Support-With_coffee!-Green\n    :target: https://www.buymeacoffee.com/licht1stein\n\n----------------------\nWhat this package does\n----------------------\nThis is a code generation package that converts YML definitions to Pydantic models (either python code or python objects).\n\n----------------\nWhat is Pydantic\n----------------\n`Pydantic <https://pydantic-docs.helpmanual.io/>`_ is a python library for data validation and settings management using\npython type annotations.\n\nTake a look at the `official example <https://pydantic-docs.helpmanual.io/#example>`_ from the Pydantic docs.\n\n---------------------\nWhy generate schemas?\n---------------------\nNormally you just program the schemas within your program, but there are several\nuse cases when code generation makes a lot of sense:\n\n- You're programming several apps that use the same schema (think an API server and client library for it)\n\n- You're programming in more than one programming language\n\n---------------\nGetting started\n---------------\n\nInstallation\n------------\nUsing pip:\n\n.. code-block:: bash\n\n    pip install pydantic-gen\n\nUsing `poetry <https://python-poetry.org/>`_:\n\n.. code-block:: bash\n\n    poetry add pydantic-gen\n\nUsage\n-----\nFirst you need to create a YAML file with your desired class schema. See `example.yml <https://github.com/licht1stein/pydantic-gen/blob/documentation/example.yml>`_ file.\n\n.. code-block:: python\n\n    from pydantic_gen import SchemaGen\n\n    generated = SchemaGen('example.yml')\n\nThe code is now generated and stored in `generated.code` attribute. There are\ntwo ways to use the code:\n\n1. Save it to a file, and use the file in your program:\n\n.. code-block:: python\n\n    generated.to_file('example_output.py')\n\nYou can inspect the resulting file in the `example_output.py <https://github.com/licht1stein/pydantic-gen/blob/documentation/example_output.py>`_\n\n2. Or directly import the generated classed directly without saving:\n\n.. code-block:: python\n\n    generated.to_sys(module_name='generated_schemas')\n\nAfter running `generated.to_sys(module_name='generated_schemas'` your generated code will be available for import:\n\n.. code-block:: python\n\n    import generated_schemas as gs\n\n    schema = gs.GeneratedSchema1(id=1)\n\nUsage pattern\n-------------\nRecommended usage pattern is creating the yaml files needed for your projects\nand storing them in a separate repository, to achieve maximum consistency across all projects.\n\nYAML-file structure\n-------------------\n`schemas` - list of all schemas described\n\n`name` - name of the generated class\n\n`props` - list of properties of the class using python type\nannotation. Fields: `name` - field name, `type` - field type,\n`optional` - bool, if True the type will be wrapped in `Optional`,\n`default` - default value for the field.\n\n`config` - list of config settings from `Model Config <https://pydantic-docs.helpmanual.io/usage/model_config/>`_\nof pydantic.\n\nTesting\n-------\nProject is fully covered by tests and uses pytest. To run:\n\n.. code-block:: bash\n\n    pytest\n\nPackaging Notice\n----------------\nThis project uses the excellent `poetry <https://python-poetry.org>`__ for packaging. Please read about it and let's all start using\n`pyproject.toml` files as a standard. Read more:\n\n* `PEP 518 -- Specifying Minimum Build System Requirements for Python Projects <https://www.python.org/dev/peps/pep-0518/>`_\n\n* `What the heck is pyproject.toml? <https://snarky.ca/what-the-heck-is-pyproject-toml/>`_\n\n* `Clarifying PEP 518 (a.k.a. pyproject.toml) <https://snarky.ca/clarifying-pep-518/>`_\n\n\n\n",
    'author': 'MB',
    'author_email': 'mb@blaster.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/licht1stein/pydantic-gen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
